"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from .detection import DetectionResult
from .policy import TYPE_POLICY, Policy
from .rule import Rule


@dataclass
class TestError:
    """Represents an error caused by any of the functions or a generic one"""

    message: Optional[str] = None


@dataclass
class FunctionTestResult:
    """Defines the result of running a function"""

    # output contains the JSON-encoded return value, None if an error was raised
    output: Optional[str]
    # error contains a TestError instance with the error message or
    # None if no error was raised
    error: Optional[TestError]
    # return value matched expectations
    matched: Optional[bool]

    @classmethod
    def new(
        cls,
        matched: bool,
        output: Optional[Union[bool, str, List[str]]],
        raw_exception: Optional[Exception] = None,
    ) -> Optional["FunctionTestResult"]:
        """Create a new instance while applying
        the necessary transformations to the parameters"""
        if output is None and raw_exception is None:
            return None

        if output is not None and not isinstance(output, str):
            output = json.dumps(output)

        return cls(output=output, error=cls.to_test_error(raw_exception), matched=matched)

    @staticmethod
    def format_exception(exc: Optional[Exception], title: Optional[str] = None) -> Optional[str]:
        """Convert an exception instance to a structured error message"""
        if exc is None:
            return None
        if title is not None:
            prefix = f"{title}: "
        else:
            prefix = ""

        return f"{prefix}{type(exc).__name__}: {exc}"

    @staticmethod
    def to_test_error(exc: Optional[Exception], title: Optional[str] = None) -> Optional[TestError]:
        """Convert an exception instance to a TestError,
        also properly formatting the exception message"""
        if exc is None:
            return None
        return TestError(message=FunctionTestResult.format_exception(exc, title=title))


@dataclass  # pylint: disable=R0902
class TestResultsPerFunction:
    """Container for the results of each function"""

    detectionFunction: Optional[FunctionTestResult]  # pylint: disable=C0103
    titleFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    dedupFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    alertContextFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    descriptionFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    referenceFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    severityFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    runbookFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    destinationsFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103


@dataclass  # pylint: disable=R0902
class TestResult:
    """The structure of the results for a test case evaluation"""

    id: Optional[str]  # pylint: disable=C0103
    name: str
    # The following two fields do not conform to Python's naming
    # conventions, but the field names correspond
    # to response attributes by API & FE.
    # TODO: provide a field name translation step if necessary
    detectionId: Optional[str]  # pylint: disable=C0103
    genericError: Optional[str]  # pylint: disable=C0103
    error: Optional[TestError]
    errored: bool
    passed: bool
    trigger_alert: Optional[bool]
    functions: TestResultsPerFunction


@dataclass
class TestExpectations:
    """Contains the expected values for performing assertions"""

    detection: bool
    # TODO: include assertions for remaining functions, e.g title and alert context


@dataclass
class TestSpecification:
    """The structure of a test case"""

    id: str  # pylint: disable=C0103
    name: str
    data: Dict[str, Any]
    mocks: List[Dict[str, Any]]
    expectations: TestExpectations


# pylint: disable=too-few-public-methods
class TestCaseEvaluator:
    """Translates detection execution results to test case results,
    by performing assertions and determining the status"""

    def __init__(self, spec: TestSpecification, detection_result: DetectionResult):
        self._spec = spec
        self._detection_result = detection_result

    def _get_result_status(self) -> bool:
        """Get the test status - passing/failing"""
        # Aux functions are executed unconditionally
        # (regardless if the detection matched or not) during testing.
        # Only if the detection is expected to trigger an alert,
        # we want to include errors from other functions in the status.
        if self._spec.expectations.detection == self._get_detection_alert_value():
            return self._detection_result.trigger_alert and not self._detection_result.errored
        # expectations match the detection output
        return self._spec.expectations.detection == self._detection_result.detection_output

    def _get_generic_error_details(self) -> Tuple[Optional[Exception], Optional[str]]:
        generic_error = None
        generic_error_title = None
        if self._detection_result.input_exception is not None:
            generic_error = self._detection_result.input_exception
            generic_error_title = "Invalid event"
        elif self._detection_result.setup_exception is not None:
            generic_error = self._detection_result.setup_exception
        return generic_error, generic_error_title

    def _get_detection_alert_value(self) -> bool:
        if self._detection_result.detection_type.upper() == TYPE_POLICY.upper():
            return Policy.matcher_alert_value
        return Rule.matcher_alert_value

    def interpret(
        self, ignore_exception_types: Optional[List[Type[Exception]]] = None
    ) -> TestResult:
        """Evaluate the detection result taking into account
        the errors raised during evaluation and
        the test specification expectations"""

        # first, we should update the detection result, taking into account any
        # ignored exception types passed into this test
        if ignore_exception_types:
            self._detection_result.ignore_errors(ignore_exception_types)

        function_results = {
            "detectionFunction": FunctionTestResult.new(
                self._spec.expectations.detection == self._detection_result.detection_output,
                self._detection_result.detection_output,
                self._detection_result.detection_exception,
            )
        }

        # We don't include output from other functions
        # unless the test was expected to match and trigger an alert.
        # If the test fails, providing all the output provides a faster feedback loop,
        # on possible additional failures.
        if self._spec.expectations.detection == self._get_detection_alert_value():
            function_results.update(
                titleFunction=FunctionTestResult.new(
                    self._detection_result.title_exception is None,
                    self._detection_result.title_output,
                    self._detection_result.title_exception,
                ),
                descriptionFunction=FunctionTestResult.new(
                    self._detection_result.description_exception is None,
                    self._detection_result.description_output,
                    self._detection_result.description_exception,
                ),
                referenceFunction=FunctionTestResult.new(
                    self._detection_result.reference_exception is None,
                    self._detection_result.reference_output,
                    self._detection_result.reference_exception,
                ),
                severityFunction=FunctionTestResult.new(
                    self._detection_result.severity_exception is None,
                    self._detection_result.severity_output,
                    self._detection_result.severity_exception,
                ),
                runbookFunction=FunctionTestResult.new(
                    self._detection_result.runbook_exception is None,
                    self._detection_result.runbook_output,
                    self._detection_result.runbook_exception,
                ),
                destinationsFunction=FunctionTestResult.new(
                    self._detection_result.destinations_exception is None,
                    self._detection_result.destinations_output,
                    self._detection_result.destinations_exception,
                ),
                dedupFunction=FunctionTestResult.new(
                    self._detection_result.dedup_exception is None,
                    self._detection_result.dedup_output,
                    self._detection_result.dedup_exception,
                ),
                alertContextFunction=FunctionTestResult.new(
                    self._detection_result.alert_context_exception is None,
                    self._detection_result.alert_context_output,
                    self._detection_result.alert_context_exception,
                ),
            )

        generic_error, generic_error_title = self._get_generic_error_details()

        return TestResult(
            id=self._spec.id,
            name=self._spec.name,
            detectionId=self._detection_result.detection_id,
            genericError=FunctionTestResult.format_exception(
                generic_error, title=generic_error_title
            ),
            errored=self._detection_result.errored,
            error=FunctionTestResult.to_test_error(generic_error, title=generic_error_title),
            # Passing or failing test?
            passed=self._get_result_status(),
            trigger_alert=self._detection_result.trigger_alert,
            functions=TestResultsPerFunction(**function_results),
        )



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import re
import traceback
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import List, Optional, Type

from .util import id_to_path, import_file_as_module, store_modules

# Temporary alias for compatibility
get_logger = logging.getLogger


# pylint: disable=too-many-instance-attributes,unsubscriptable-object
@dataclass
class DetectionResult:
    """Class containing the result of running a detection"""

    detection_id: str
    detection_severity: str
    detection_type: str
    trigger_alert: bool  # detection output, default to non-alerting value

    setup_exception: Optional[Exception] = None

    detection_output: Optional[bool] = None
    detection_exception: Optional[Exception] = None

    filter_output: Optional[bool] = None
    filter_exception: Optional[Exception] = None

    dedup_output: Optional[str] = None
    dedup_exception: Optional[Exception] = None
    dedup_defined: bool = False

    title_output: Optional[str] = None
    title_exception: Optional[Exception] = None
    title_defined: bool = False

    description_output: Optional[str] = None
    description_exception: Optional[Exception] = None
    description_defined: bool = False

    reference_output: Optional[str] = None
    reference_exception: Optional[Exception] = None
    reference_defined: bool = False

    severity_output: Optional[str] = None
    severity_exception: Optional[Exception] = None
    severity_defined: bool = False

    runbook_output: Optional[str] = None
    runbook_exception: Optional[Exception] = None
    runbook_defined: bool = False

    destinations_output: Optional[List[str]] = None
    destinations_exception: Optional[Exception] = None
    destinations_defined: bool = False

    alert_context_output: Optional[str] = None
    alert_context_exception: Optional[Exception] = None
    alert_context_defined: bool = False

    input_exception: Optional[Exception] = None

    @property
    def fatal_error(self) -> Optional[Exception]:
        """Provide any error that would stop evaluation
        or None, if no blocking error is found"""
        exception = None
        if self.setup_exception:
            exception = self.setup_exception
        elif self.detection_exception:
            exception = self.detection_exception
        elif self.filter_exception:
            exception = self.filter_exception
        elif self.input_exception:
            exception = self.input_exception
        return exception

    @property
    def error_type(self) -> Optional[str]:
        """Returns the type of the exception, None if there was no error"""
        fatal_error = self.fatal_error
        if fatal_error is None:
            return None
        return type(fatal_error).__name__

    @property
    def short_error_message(self) -> Optional[str]:
        """Returns short error message, None if there was no error"""
        fatal_error = self.fatal_error
        if fatal_error is None:
            return None
        return repr(fatal_error)

    @property
    def error_message(self) -> Optional[str]:
        """Returns formatted error message with traceback"""
        exception = self.fatal_error
        if exception is None:
            return None

        trace = traceback.format_tb(exception.__traceback__)

        if len(trace) == 0:
            # an exception might not have a trace back if it was not raised
            return str(exception)

        # we only take last element of trace which will show the
        # detection file name and line of the error, for example:
        #    division by zero: AlwaysFail.py, line 4, in detection 1/0
        file_trace = trace[len(trace) - 1].strip().replace("\n", "")
        # this looks like: File "/tmp/rules/AlwaysFail.py", line 4, in detection 1/0 BUT
        # we want just the file name
        return str(exception) + ": " + re.sub(r'File.*/(.*[.]py)"', r"\1", file_trace)

    @property
    def errored(self) -> bool:
        """Returns whether any of the detection functions raised an error"""
        return bool(
            self.detection_exception
            or self.filter_exception
            or self.title_exception
            or self.dedup_exception
            or self.alert_context_exception
            or self.description_exception
            or self.reference_exception
            or self.severity_exception
            or self.runbook_exception
            or self.destinations_exception
            or self.setup_exception
            or self.input_exception
        )

    @property
    def detection_evaluation_failed(self) -> bool:
        """Returns whether the detection/filter functions raise an error or an import error occurred"""
        return bool(self.detection_exception or self.filter_exception or self.setup_exception)

    def ignore_errors(self, ignore_exception_types: List[Type[Exception]]) -> None:
        """Used to ignore exceptions of particular types, used primarily in testing"""
        for exception_type in ignore_exception_types:
            if isinstance(self.detection_exception, exception_type):
                self.detection_exception = None
            if isinstance(self.filter_exception, exception_type):
                self.filter_exception = None
            if isinstance(self.title_exception, exception_type):
                self.title_exception = None
            if isinstance(self.description_exception, exception_type):
                self.description_exception = None
            if isinstance(self.reference_exception, exception_type):
                self.reference_exception = None
            if isinstance(self.severity_exception, exception_type):
                self.severity_exception = None
            if isinstance(self.runbook_exception, exception_type):
                self.runbook_exception = None
            if isinstance(self.destinations_exception, exception_type):
                self.destinations_exception = None
            if isinstance(self.dedup_exception, exception_type):
                self.dedup_exception = None
            if isinstance(self.alert_context_exception, exception_type):
                self.alert_context_exception = None

    @property
    def all_filters_matched(self) -> bool:
        return bool(self.filter_output and not self.filter_exception)

    @property
    def filters_did_not_match(self) -> bool:
        return (
            self.filter_output is not None and not self.filter_output and not self.filter_exception
        )

    @property
    def filters_not_ran(self) -> bool:
        return self.filter_output is None and not self.filter_exception


class BaseImporter:
    """Base class for Python module importers"""

    @staticmethod
    def from_file(identifier: str, path: str) -> ModuleType:
        """Import a file as a Python module"""
        return import_file_as_module(path, identifier)

    def from_string(self, identifier: str, body: str, tmp_dir: str) -> ModuleType:
        """Write source code to a temporary file and import as Python module"""
        path = id_to_path(tmp_dir, identifier)
        store_modules(path, body)
        return self.from_file(identifier, path)

    @abstractmethod
    def get_module(self, identifier: str, resource: str) -> ModuleType:
        pass


class FilesystemImporter(BaseImporter):
    """Import a Python module from the filesystem"""

    def get_module(  # pylint: disable=arguments-differ,W0237
        self, identifier: str, path: str
    ) -> ModuleType:
        module_path = Path(path)
        if not module_path.exists():
            raise FileNotFoundError(path)
        return super().from_file(identifier, module_path.absolute().as_posix())


class RawStringImporter(BaseImporter):
    """Import a Python module from raw source code"""

    def __init__(self, tmp_dir: str):
        super().__init__()
        self._tmp_dir = tmp_dir

    def get_module(  # pylint: disable=arguments-differ,W0237
        self, identifier: str, body: str
    ) -> ModuleType:
        return super().from_string(identifier, body, self._tmp_dir)



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from ..policy import ERROR_TYPE_POLICY
from ..rule import (
    ERROR_TYPE_CORRELATION_RULE,
    ERROR_TYPE_RULE,
    ERROR_TYPE_SCHEDULED_RULE,
)

# Aliases
ExecutionInputData = Dict[str, Any]
ExecutionEnvComponent = Dict[str, Any]

LogEventInput = Dict[str, Any]
CloudResourceInput = Dict[str, Any]


@dataclass(frozen=True)
class ExecutionMatch:
    # required for all matches
    alert_type: str
    detection_type: str
    detection_id: str
    detection_version: str
    detection_tags: List[str]
    detection_reports: Dict[str, List[str]]
    detection_severity: str
    dedup_string: str
    dedup_period_mins: int
    event: Dict[str, Any]
    # some of these will be set
    event_id: Optional[str] = None
    replay_id: Optional[str] = None
    event_timestamp: Optional[str] = None  # Use ISO 8601! %Y-%m-%dT%H:%M:%S.%f%z
    # optional dynamic fields
    severity: Optional[str] = None
    alert_context: Optional[str] = None
    description: Optional[str] = None
    destinations: Optional[List[str]] = None
    reference: Optional[str] = None
    runbook: Optional[str] = None
    title: Optional[str] = None

    @property
    def errored(self) -> bool:
        return self.alert_type in (
            ERROR_TYPE_RULE,
            ERROR_TYPE_SCHEDULED_RULE,
            ERROR_TYPE_POLICY,
            ERROR_TYPE_CORRELATION_RULE,
        )

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> Optional[ExecutionMatch]:
        if data is not None:
            return cls(**data)
        return None


@dataclass(frozen=True)
class _BaseDataObject:
    def to_json(self) -> str:
        return json.dumps(asdict(self))

    def to_bytes(self) -> bytes:
        return self.to_json().encode("utf-8")


@dataclass(frozen=True)
class ClientOptions(_BaseDataObject):
    lambda_name: str


class ExecutionMode(str, Enum):
    S3 = "S3"
    NONE = "NONE"
    INLINE = "INLINE"


class ExecutionStatus(str, Enum):
    OK = "OK"
    ERROR = "ERROR"
    TIMEOUT = "TIMEOUT"



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from .common import (
    CloudResourceInput,
    ExecutionEnvComponent,
    ExecutionInputData,
    ExecutionMode,
    LogEventInput,
    _BaseDataObject,
)


@dataclass(frozen=True)
class ExecutionTaskInput(_BaseDataObject):
    _event_input_id = "p_row_id"
    _resource_input_id = "resourceId"

    mode: ExecutionMode
    data: List[ExecutionInputData]
    input_id_field: str
    url: Optional[str] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionTaskInput:
        return cls(
            mode=ExecutionMode(data["mode"]),
            url=data.get("url"),
            data=data["data"],
            input_id_field=data["input_id_field"],
        )

    @classmethod
    def inline_resources(cls, resources: List[CloudResourceInput]) -> ExecutionTaskInput:
        return cls(
            url=None,
            mode=ExecutionMode.INLINE,
            data=resources,
            input_id_field=cls._resource_input_id,
        )

    @classmethod
    def inline_events(cls, events: List[LogEventInput]) -> ExecutionTaskInput:
        return cls(
            url=None,
            mode=ExecutionMode.INLINE,
            data=events,
            input_id_field=cls._event_input_id,
        )


@dataclass(frozen=True)
class ExecutionTaskOutput(_BaseDataObject):
    mode: ExecutionMode
    url: Optional[str] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionTaskOutput:
        return cls(
            mode=ExecutionMode(data["mode"]),
            url=data.get("url"),
        )

    @classmethod
    def inline(cls) -> ExecutionTaskOutput:
        return cls(
            url=None,
            mode=ExecutionMode.INLINE,
        )


@dataclass(frozen=True)
class ExecutionTaskOptions(_BaseDataObject):
    execution_details: bool
    timeout_seconds: Optional[int] = None
    metadata: Optional[Dict[str, str]] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionTaskOptions:
        return cls(
            execution_details=data["execution_details"],
            timeout_seconds=data.get("timeout_seconds"),
            metadata=data.get("metadata"),
        )


@dataclass(frozen=True)
class ExecutionEnv(_BaseDataObject):
    mocks: List[ExecutionEnvComponent]
    outputs: List[ExecutionEnvComponent]
    globals: List[ExecutionEnvComponent]
    detections: List[ExecutionEnvComponent]
    data_model: Optional[ExecutionEnvComponent]

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionEnv:
        return cls(
            mocks=data.get("mocks", []),
            outputs=data.get("outputs", []),
            globals=data.get("globals", []),
            detections=data.get("detections", []),
            data_model=data.get("data_model"),
        )


@dataclass(frozen=True)
class ExecutionTaskEnv(_BaseDataObject):
    mode: ExecutionMode
    url: Optional[str] = None
    env: Optional[ExecutionEnv] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionTaskEnv:
        return cls(
            mode=ExecutionMode(data["mode"]),
            url=data.get("url"),
            env=ExecutionEnv.from_json(data.get("env", {})),
        )

    @classmethod
    def inline(cls, env: ExecutionEnv) -> ExecutionTaskEnv:
        return cls(
            env=env,
            url=None,
            mode=ExecutionMode.INLINE,
        )


@dataclass(frozen=True)
class ExecutionTask(_BaseDataObject):
    env: ExecutionTaskEnv
    input: ExecutionTaskInput
    output: ExecutionTaskOutput
    options: ExecutionTaskOptions

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionTask:
        return cls(
            env=ExecutionTaskEnv.from_json(data["env"]),
            input=ExecutionTaskInput.from_json(data["input"]),
            output=ExecutionTaskOutput.from_json(data["output"]),
            options=ExecutionTaskOptions.from_json(data["options"]),
        )



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import OrderedDict
from collections.abc import Mapping
from functools import reduce
from typing import Any, List, Optional, Sequence, Union

from dateutil import parser as dateparser

from .data_model import E_NO_DATA_MODEL_FOUND, DataModel
from .exceptions import PantherError
from .immutable import ImmutableCaseInsensitiveDict, json_encoder

__all__ = ["PantherEvent"]


class PantherEvent(ImmutableCaseInsensitiveDict):  # pylint: disable=R0901
    """Panther enriched event with unified data model (udm) access."""

    def __init__(self, event: Mapping, data_model: Optional[DataModel] = None):
        """Create data model lookups

        Args:
            event: Dictionary representing the event.
            data_model: the data model used for the LogType associated with this event
        """
        super().__init__(event)
        self.data_model = data_model

    def udm(self, *key: str, default: Any = None) -> Any:
        """
        udm operates in two modes

        Mode 1: Data Model Access
        If a single key is provided it will check to see if it exists on the data
        model mapping. If a mapping is not found it operates in Mode 2.

        Mode 2: p_udm access
        If there was no match against the data model it checks against the p_udm field.
        If there was no match against the p_udm field it will check on the event itself.
        If all other conditions are exhausted the default value is utilized as the return value
        """

        model = self.data_model
        if len(key) == 1 and model:
            match = self._get_json_path(key[0])
            if match:
                return self._ensure_immutable(match.value)
            method = self._get_method(key[0])
            if method:
                return self._ensure_immutable(method(self._ensure_immutable(self._container)))

            # if data model mapping is defined we should short circuit logic
            # to maintain backwards compatability
            if key[0] in model.paths or key[0] in model.methods:
                return default

        pieces = list(key)
        if len(pieces) > 0:
            pieces.insert(0, "p_udm")
            result = self.deep_get(*tuple(pieces))
            if not result:
                result = self.deep_get(*list(key))
            if result:
                return result
        # no matches, return default
        return default

    def udm_path(self, key: str) -> Optional[str]:
        """Returns the JSON path or method name for the mapped field"""
        self._validate()
        # access values via standardized fields
        match = self._get_json_path(key)
        if match:
            return str(match.full_path)
        method = self._get_method(key)
        if method:
            return getattr(method, "__name__", repr(method))
        # no matches, return None by default
        return None

    def deep_get(self, *keys: str, default: Any = None) -> Any:
        """Safely return the value of an arbitrarily nested map"""
        out = reduce(
            lambda d, key: d.get(key, default) if isinstance(d, Mapping) else default,
            keys,
            self,
        )
        if out is None:
            return default
        return out

    def deep_walk(
        self, *keys: str, default: Optional[str] = None, return_val: str = "all"
    ) -> Union[Optional[Any], Optional[List[Any]]]:
        """Safely retrieve a value stored in complex dictionary structure

        Similar to deep_get but supports accessing dictionary keys within nested lists as well

        Parameters:
        keys (str): comma-separated list of keys used to traverse the event object
        default (str): the default value to return if the desired key's value is not present
        return_val (str): string specifying which value to return
                        possible values are "first", "last", or "all"

        Returns:
        any | list[any]: A single value if return_val is "first", "last",
                        or if "all" is a list containing one element,
                        otherwise a list of values
        """

        def _empty_list(sub_obj: Any) -> bool:
            return (
                all(_empty_list(next_obj) for next_obj in sub_obj)
                if isinstance(sub_obj, Sequence) and not isinstance(sub_obj, str)
                else False
            )

        obj = self._container

        if not keys:
            return default if _empty_list(obj) or obj is None else obj

        current_key = keys[0]
        found: OrderedDict = OrderedDict()

        if isinstance(obj, Mapping):
            next_key = PantherEvent(obj.get(current_key, default))
            return (
                next_key.deep_walk(*keys[1:], default=default, return_val=return_val)
                if next_key is not None
                else default
            )
        if isinstance(obj, Sequence) and not isinstance(obj, str):
            for item in obj:
                next_item = PantherEvent(item)
                value = next_item.deep_walk(*keys, default=default, return_val=return_val)
                if value is not None:
                    if isinstance(value, Sequence) and not isinstance(value, str):
                        for sub_item in value:
                            found[sub_item] = None
                    else:
                        found[value] = None
        found_list: list[Any] = list(found.keys())
        if not found_list:
            return default
        return {
            "first": found_list[0],
            "last": found_list[-1],
            "all": found_list[0] if len(found_list) == 1 else found_list,
        }.get(return_val, "all")

    def event_time_epoch(self, force_utc: bool = True) -> int:
        """Parses epoch seconds from the p_event_time field.
        Returns 0 if event time is not available."""
        try:
            event_time = self.get("p_event_time") or ""
            if force_utc:
                event_time += " UTC"
            return int(dateparser.parse(event_time).timestamp())
        except dateparser.ParserError:
            return 0

    def _validate(self) -> None:
        if not self.data_model:
            raise PantherError(E_NO_DATA_MODEL_FOUND, self._container.get("p_log_type"))

    def _get_json_path(self, key: str) -> Any:
        if not self.data_model:  # makes linter happy, we never call this if not set
            return None
        if key not in self.data_model.paths:
            return None
        json_path = self.data_model.paths.get(key)
        if not json_path:
            return None
        matches = json_path.find(self._container)
        if len(matches) == 0:
            return None
        if len(matches) == 1:
            return matches[0]
        # pylint:disable=broad-exception-raised
        raise Exception(
            "JSONPath [{}] in DataModel [{}], matched multiple fields.".format(
                json_path, self.data_model.data_model_id
            )
        )

    def _get_method(self, key: str) -> Any:
        if not self.data_model:  # makes linter happy, we never call this if not set
            return None
        if key not in self.data_model.methods:
            return None
        method = self.data_model.methods.get(key)
        if callable(method):
            return method
        # no matches, return None by default
        return None

    json_encoder = json_encoder



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from .common import ExecutionMatch, ExecutionMode, ExecutionStatus, _BaseDataObject


@dataclass(frozen=True)
class ExecutionPrimaryFunctionDetails(_BaseDataObject):
    error: Optional[str] = None
    output: Optional[bool] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionPrimaryFunctionDetails:
        return cls(
            error=data.get("error"),
            output=data.get("output"),
        )


@dataclass(frozen=True)
class ExecutionDetailsPrimaryFunctions(_BaseDataObject):
    detection: ExecutionPrimaryFunctionDetails

    @property
    def errored(self) -> bool:
        return self.detection.error is not None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionDetailsPrimaryFunctions:
        return cls(
            detection=ExecutionPrimaryFunctionDetails.from_json(data.get("detection", {})),
        )


@dataclass(frozen=True)
class ExecutionAuxFunctionDetails(_BaseDataObject):
    defined: bool
    error: Optional[str] = None
    output: Optional[str] = None

    @property
    def errored(self) -> bool:
        return self.error is not None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionAuxFunctionDetails:
        return cls(
            defined=data["defined"],
            error=data["error"],
            output=data["output"],
        )


@dataclass(frozen=True)
class ExecutionDetailsAuxFunctions(_BaseDataObject):
    dedup: ExecutionAuxFunctionDetails
    title: ExecutionAuxFunctionDetails
    runbook: ExecutionAuxFunctionDetails
    severity: ExecutionAuxFunctionDetails
    reference: ExecutionAuxFunctionDetails
    description: ExecutionAuxFunctionDetails
    destinations: ExecutionAuxFunctionDetails
    alert_context: ExecutionAuxFunctionDetails

    @property
    def errored(self) -> bool:
        return (
            self.title.errored
            or self.runbook.errored
            or self.severity.errored
            or self.reference.errored
            or self.description.errored
            or self.destinations.errored
            or self.dedup.errored
            or self.alert_context.errored
        )

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionDetailsAuxFunctions:
        return cls(
            dedup=ExecutionAuxFunctionDetails.from_json(data["dedup"]),
            title=ExecutionAuxFunctionDetails.from_json(data["title"]),
            runbook=ExecutionAuxFunctionDetails.from_json(data["runbook"]),
            severity=ExecutionAuxFunctionDetails.from_json(data["severity"]),
            reference=ExecutionAuxFunctionDetails.from_json(data["reference"]),
            description=ExecutionAuxFunctionDetails.from_json(data["description"]),
            destinations=ExecutionAuxFunctionDetails.from_json(data["destinations"]),
            alert_context=ExecutionAuxFunctionDetails.from_json(data["alert_context"]),
        )


@dataclass(frozen=True)
class ExecutionDetails(_BaseDataObject):
    aux_functions: ExecutionDetailsAuxFunctions
    primary_functions: ExecutionDetailsPrimaryFunctions
    input_error: Optional[str] = None
    setup_error: Optional[str] = None

    @property
    def errored(self) -> bool:
        return (
            self.input_error is not None
            or self.setup_error is not None
            or self.aux_functions.errored
            or self.primary_functions.errored
        )

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionDetails:
        return cls(
            input_error=data.get("input_error"),
            setup_error=data.get("setup_error"),
            aux_functions=ExecutionDetailsAuxFunctions.from_json(data["aux_functions"]),
            primary_functions=ExecutionDetailsPrimaryFunctions.from_json(data["primary_functions"]),
        )


@dataclass(frozen=True)
class ExecutionOutput(_BaseDataObject):
    input_id: str
    match: Optional[ExecutionMatch] = None
    details: Optional[ExecutionDetails] = None

    @property
    def trigger_alert(self) -> bool:
        return self.match is not None or self.errored is True

    @property
    def errored(self) -> bool:
        return bool((self.match and self.match.errored) or (self.details and self.details.errored))

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionOutput:
        return cls(
            match=ExecutionMatch.from_json(data["match"]),
            details=ExecutionDetails.from_json(data["details"]),
            input_id=data["input_id"],
        )


@dataclass(frozen=True)
class ExecutionStats(_BaseDataObject):
    total_execution_ns: int = 0
    avg_execution_ns: float = 0
    total_events: int = 0
    last_run_ns: int = 0

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionStats:
        return cls(
            total_execution_ns=data.get("total_execution_ns", 0),
            avg_execution_ns=data.get("avg_execution_ns", 0),
            total_events=data.get("total_events", 0),
            last_run_ns=data.get("last_run_ns", 0),
        )


@dataclass(frozen=True)
class ExecutionResult(_BaseDataObject):
    output_mode: ExecutionMode
    status: Optional[ExecutionStatus] = None
    url: Optional[str] = None
    data: Optional[List[ExecutionOutput]] = None
    metadata: Optional[Dict[str, str]] = None
    detection_stats: Optional[Dict[str, ExecutionStats]] = None

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> ExecutionResult:
        output_mode = data.get("output_mode")
        if output_mode is None:
            raise TypeError(f"Expected ExecutionResult, but got: {data}")

        outputs = []
        for json_str in data.get("data") or []:
            obj = ExecutionOutput.from_json(json_str)
            outputs.append(obj)

        status: Optional[ExecutionStatus] = None
        data_status = data.get("status")
        if data_status is not None:
            status = ExecutionStatus(data_status)
        detection_stats: Optional[Dict[str, ExecutionStats]] = None
        raw_stats: Dict[str, Any]
        if data.get("detection_stats") is not None:
            detection_stats = {}
            raw_stats = data.get("detection_stats", {})
            for detection_id, value in raw_stats.items():
                detection_stats[detection_id] = ExecutionStats.from_json(value)

        return cls(
            url=data.get("url"),
            status=status,
            detection_stats=detection_stats,
            data=outputs,
            metadata=data.get("metadata"),
            output_mode=ExecutionMode(output_mode),
        )



import ast
import inspect
import logging
import textwrap
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

__all__ = ["FuncTransformer", "snapshot_func"]


class FuncTransformer(ast.NodeTransformer):
    # pylint: disable=invalid-name,broad-except,eval-used

    def __init__(self, ns: Dict[Any, Any], local_names: Optional[List[str]] = None):
        self._ns = ns
        self._local_names = local_names or []
        self.errors: List[BaseException] = []

    # remove type annotations and decorators
    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        try:
            self._local_names.append(node.name)

            node.returns = None
            if node.args.args:
                for arg in node.args.args:
                    self._local_names.append(arg.arg)
                    arg.annotation = None

            node.decorator_list.clear()

            return self._closure_delve(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    # read imports into _local_names
    def visit_Import(self, node: ast.Import) -> ast.AST:
        try:
            for name in node.names:
                self._local_names.append(name.asname or name.name)

            return self.generic_visit(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    # read from-imports into _local_names
    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.AST:
        try:
            for name in node.names:
                self._local_names.append(name.asname or name.name)

            return self.generic_visit(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    # auto-resolve attributes to constant values
    def visit_Attribute(self, node: ast.Attribute) -> ast.AST:
        try:
            root_name, skip = self._attribute_root(node)
            if skip:
                return self.generic_visit(node)

            if isinstance(node.ctx, ast.Load) and (root_name not in self._local_names):
                return self._const_replace(node)

            return self.generic_visit(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    # auto-resolve names to constant values
    def visit_Name(self, node: ast.Name) -> ast.AST:
        try:
            if isinstance(node.ctx, ast.Store):
                self._local_names.append(node.id)
                return self.generic_visit(node)

            if isinstance(node.ctx, ast.Load) and (node.id not in self._local_names):
                return self._const_replace(node)

            return self.generic_visit(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    def visit_Lambda(self, node: ast.Lambda) -> ast.AST:
        try:
            return self._closure_delve(node)

        except BaseException as err:
            return self._report_error_and_skip(node, err)

    def _closure_delve(self, node: Union[ast.Lambda, ast.FunctionDef]) -> ast.AST:
        prev_args = self._local_names.copy()
        for arg in node.args.args:
            self._local_names.append(arg.arg)

        new_node = self.generic_visit(node)
        self._local_names = prev_args

        return new_node

    def _const_replace(self, node: ast.AST) -> Union[ast.Constant, ast.AST]:
        try:
            c = ast.unparse(node)  # nosec B307
            x = eval(c, self._ns)  # nosec B307

            const_node = ast.Constant(x)

            if inspect.isbuiltin(x):
                return self.generic_visit(node)

            if ast.unparse(cast(ast.AST, const_node))[0] == "<":
                return self.generic_visit(node)

            return const_node
        except BaseException as err:
            return self._report_error_and_skip(node, err)

    def _attribute_root(self, node: Union[ast.AST, ast.expr]) -> Tuple[str, bool]:
        if isinstance(node, ast.Name):
            return node.id, False

        if isinstance(node, (ast.Attribute, ast.Subscript, ast.Constant)):
            return self._attribute_root(node.value)

        if isinstance(node, ast.Call):
            return self._attribute_root(node.func)

        # else, it is a complex case, process children individually
        return "", True

    def _report_error_and_skip(self, node: ast.AST, err: BaseException) -> ast.AST:
        node_code = ast.unparse(node)
        logging.warning("unable to parse: %s", node_code)
        logging.error(err)
        self.errors.append(err)
        return self.generic_visit(node)


def snapshot_func(target_func: Callable) -> Tuple[str, List[BaseException]]:
    original_src = textwrap.dedent(inspect.getsource(target_func))
    original_tree = ast.parse(
        original_src,
        filename=(inspect.getsourcefile(target_func) or "unknown"),
        mode="exec",
    )

    v_nonlocal, v_global, _, _ = inspect.getclosurevars(target_func)
    transformer = FuncTransformer(ns={**v_nonlocal, **v_global})

    parsed_tree = transformer.visit(original_tree)
    parsed_src = ast.unparse(parsed_tree)

    return parsed_src, transformer.errors



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections.abc import Mapping

from .rule import Detection

TYPE_POLICY = "POLICY"
ERROR_TYPE_POLICY = "POLICY_ERROR"


class Policy(Detection):
    """Panther policy metadata and imported module."""

    # default detection types for policies
    default_detection_type = TYPE_POLICY

    # policies have a rule method
    matcher_function_name = "policy"

    # a policy should trigger an alert on False return value
    matcher_alert_value = False

    def matcher_function(self, event: Mapping) -> bool:
        command = getattr(self._module, self.matcher_function_name)
        return self._run_command(command, event, bool)



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import contextlib
import json
import logging
import os
import tempfile
from abc import ABC, abstractmethod
from collections.abc import Mapping
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional

from .detection import (
    BaseImporter,
    DetectionResult,
    FilesystemImporter,
    RawStringImporter,
)
from .enriched_event import PantherEvent
from .exceptions import FunctionReturnTypeError, UnknownDestinationError
from .filter import Filter
from .util import get_bool_env_var

# We want to default this to false as PAT will want detection output
DISABLE_OUTPUT = get_bool_env_var("DISABLE_DETECTION_OUTPUT", False)

# Temporary alias for compatibility
get_logger = logging.getLogger

TYPE_RULE = "RULE"
TYPE_SCHEDULED_RULE = "SCHEDULED_RULE"
TYPE_CORRELATION_RULE = "CORRELATION_RULE"

ERROR_TYPE_RULE = "RULE_ERROR"
ERROR_TYPE_SCHEDULED_RULE = "SCHEDULED_RULE_ERROR"
ERROR_TYPE_CORRELATION_RULE = "CORRELATION_RULE_ERROR"

_DETECTION_FOLDER = os.path.join(tempfile.gettempdir(), "detections")

# Maximum size for a dedup string
MAX_DEDUP_STRING_SIZE = 1000

# Maximum size for a generated field
MAX_GENERATED_FIELD_SIZE = 1000

# Maximum number of destinations
MAX_DESTINATIONS_SIZE = 10

# The limit for DDB is 400kb per item (we store this one in DDB) and
# the limit for SQS/SNS is 256KB. The limit of 200kb is an approximation - the other
# fields included in the request will be less than the remaining 56kb
MAX_ALERT_CONTEXT_SIZE = 200 * 1024  # 200kb

ALERT_CONTEXT_ERROR_KEY = "_error"

TRUNCATED_STRING_SUFFIX = "... (truncated)"

DEFAULT_DETECTION_DEDUP_PERIOD_MINS = 60

# Used to check dynamic severity output
SEVERITY_TYPES = ["INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"]
SEVERITY_DEFAULT = "DEFAULT"

ALERT_CONTEXT_FUNCTION = "alert_context"
DEDUP_FUNCTION = "dedup"
DESCRIPTION_FUNCTION = "description"
DESTINATIONS_FUNCTION = "destinations"
REFERENCE_FUNCTION = "reference"
RUNBOOK_FUNCTION = "runbook"
SEVERITY_FUNCTION = "severity"
TITLE_FUNCTION = "title"

# Auxiliary functions are optional
AUXILIARY_FUNCTIONS = (
    ALERT_CONTEXT_FUNCTION,
    DEDUP_FUNCTION,
    DESCRIPTION_FUNCTION,
    DESTINATIONS_FUNCTION,
    REFERENCE_FUNCTION,
    RUNBOOK_FUNCTION,
    SEVERITY_FUNCTION,
    TITLE_FUNCTION,
)


# pylint: disable=too-many-instance-attributes
class Detection(ABC):
    """Panther detection metadata and imported module."""

    # pylint: disable=too-many-branches,too-many-statements
    def __init__(self, config: Mapping):
        """Create new detection from a dict.
        Args:
            config: Dictionary that we expect to have the following keys:
                analysisType: either RULE, SCHEDULED_RULE, or POLICY
                id: Unique detection identifier
                body: The detection body
                versionId: The version of the detection
                (Optional) path: The detection module path
                (Optional) dedupPeriodMinutes: The period during which
                the events will be deduplicated
        """
        self.logger = get_logger()
        self.suppress_output = DISABLE_OUTPUT

        # Check for required string fields
        for each_field in ["id", "versionId"]:
            if not (each_field in config) or not isinstance(config[each_field], str):
                raise AssertionError('Field "%s" of type str is required field' % each_field)

        if not (config.get("body") or config.get("path")):
            raise ValueError('one of "body", "path" must be defined')

        if config.get("body") and config.get("path"):
            raise ValueError('only one of "body", "path" must be defined')

        if not any((isinstance(config.get("body"), str), isinstance(config.get("path"), str))):
            raise TypeError('"body", "path" parameters must be of string type')

        # backwards compatible for data from before we returned this
        self.detection_type = config.get("analysisType", self.default_detection_type)

        self.detection_id = config["id"]
        self.detection_version = config["versionId"]

        # TODO: severity and other detection metadata is not passed through when we run tests.
        #       https://app.asana.com/0/1200360676535738/1200403272293475
        if "severity" in config:
            self.detection_severity = config["severity"]
        else:
            self.detection_severity = None

        if not ("dedupPeriodMinutes" in config) or not isinstance(
            config["dedupPeriodMinutes"], int
        ):
            self.detection_dedup_period_mins = DEFAULT_DETECTION_DEDUP_PERIOD_MINS
        else:
            self.detection_dedup_period_mins = config["dedupPeriodMinutes"]

        if not ("tags" in config) or not isinstance(config["tags"], list):
            self.detection_tags: List[str] = []
        else:
            config["tags"].sort()
            self.detection_tags = config["tags"]

        if "reports" not in config or not config.get("reports"):
            self.detection_reports: Dict[str, List[str]] = {}
        else:
            # Reports are Dict[str, List[str]]
            # Sorting the List before setting it
            for values in config["reports"].values():
                values.sort()
            self.detection_reports = config["reports"]

        self.detection_suppressions = config.get("suppressions", [])

        self._setup_exception = None
        try:
            self._module = self._load_detection(self.detection_id, config)
            if not self._is_function_defined(self.matcher_function_name):
                raise AssertionError(
                    f"detection needs to have a method named '{self.matcher_function_name}'"
                )
        except Exception as err:  # pylint: disable=broad-except
            self._setup_exception = err
            return

        self._default_dedup_string = "defaultDedupString:{}".format(self.detection_id)
        self._auxiliary_function_definitions = self._check_defined_functions()

        filts = config.get("filters")
        if filts is None:
            self.filters: List[Filter] = []
        elif isinstance(filts, dict):
            if filts.get("statement") is None:
                raise TypeError("filter does not contain required 'statement' attribute")
            self.filters = [Filter(filts["statement"])]
        elif isinstance(filts, list) and all(isinstance(f, dict) for f in filts):
            if any(f.get("statement") is None for f in filts):
                raise TypeError("filter does not contain required 'statement' attribute")
            self.filters = [Filter(f["statement"]) for f in filts]
        elif isinstance(filts, Filter):
            self.filters = [filts]
        elif isinstance(filts, list) and all(isinstance(f, Filter) for f in filts):
            self.filters = filts
        else:
            raise TypeError(
                "expected 'filters' to be of type Union[List[dict], dict, List[Filter], Filter, None]"
            )

        self.skip_filters: bool = bool(config.get("skip_filters"))

        if not ("suppressAlert" in config) or not isinstance(config["suppressAlert"], bool):
            self.suppress_alert = False
        else:
            self.suppress_alert = config["suppressAlert"]

        self.correlation_rule_by_field_name_matches = config.get(
            "correlationRuleByFieldNameMatches", {}
        )

    @staticmethod
    def _load_detection(identifier: str, config: Mapping) -> ModuleType:
        """Load the detection code as Python module.
        Code can be provided as raw string or a path in the local filesystem.
        """
        has_raw_code = bool(config.get("body"))

        importer: Optional[BaseImporter] = None
        if has_raw_code:
            resource = config["body"]
            importer = RawStringImporter(tmp_dir=_DETECTION_FOLDER)
        else:
            resource = config["path"]
            importer = FilesystemImporter()

        return importer.get_module(identifier, resource)

    def _check_defined_functions(self) -> Dict[str, bool]:
        function_definitions = {}
        for name in AUXILIARY_FUNCTIONS:
            function_definitions[name] = self._is_function_defined(name)
        return function_definitions

    @property
    @abstractmethod
    def default_detection_type(self) -> str:
        pass

    @abstractmethod
    def matcher_function(self, event: Mapping) -> bool:
        pass

    @property
    @abstractmethod
    def matcher_function_name(self) -> str:
        pass

    @property
    @abstractmethod
    def matcher_alert_value(self) -> bool:
        pass

    @property
    def module(self) -> Any:
        """Used to expose the loaded python module to the engine,
        solely added in to support unit test mocking"""
        return self._module

    @property
    def setup_exception(self) -> Any:
        """Used to expose the setup exception to the engine"""
        return self._setup_exception

    @setup_exception.setter
    def setup_exception(self, val: Any) -> Any:
        """Used by the engine to set _setup_exception"""
        self._setup_exception = val

    def run(
        self, event: Mapping, outputs: dict, outputs_names: dict, batch_mode: bool = True
    ) -> DetectionResult:
        """
        Analyze a log line with this detection and return True, False, or an error.
        :param event: The event to run the detection against
        :param outputs: Destinations loaded from the panther-outputs-api
        :param outputs_names: Destinations mapped by their display name
        :param batch_mode: Whether the detection runs as part of the log analysis
        or as part of a simple detection test. In batch mode, title/dedup functions
        are not checked if the detection won't trigger an alert and also title()/dedup()
        won't raise exceptions, so that an alert won't be missed.
        """
        detection_result = DetectionResult(
            detection_id=self.detection_id,
            detection_severity=self.detection_severity,
            detection_type=self.detection_type,
            # set default to not alert
            trigger_alert=False,
        )
        # If there was an error setting up the detection
        # return early
        if self._setup_exception:
            detection_result.setup_exception = self._setup_exception
            return detection_result

        try:
            # Run filters before running matcher function
            if self._filters_should_run():
                # all filters must return True in order to run the match function (match func is basically a filter)
                detection_result.filter_output = all(f.filter(event) for f in self.filters)
        except Exception as err:  # pylint: disable=broad-except
            detection_result.filter_exception = err
            self.logger.info("filters errored on detection %s", self.detection_id, exc_info=err)

        try:
            if detection_result.filters_did_not_match:
                detection_result.detection_output = False
            elif detection_result.filters_not_ran or detection_result.all_filters_matched:
                detection_result.detection_output = self.matcher_function(event)
        except Exception as err:  # pylint: disable=broad-except
            detection_result.detection_exception = err

        detection_result.trigger_alert = (
            detection_result.detection_output is self.matcher_alert_value
            and (detection_result.all_filters_matched or detection_result.filters_not_ran)
        )

        if batch_mode and not detection_result.trigger_alert:
            # In batch mode (log analysis), there is no need to run the title/dedup functions
            # if the detection isn't going to trigger an alert
            return detection_result

        try:
            detection_result.title_defined = self._auxiliary_function_definitions[TITLE_FUNCTION]
            if detection_result.title_defined:
                detection_result.title_output = self._get_title(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.title_exception = err

        try:
            detection_result.description_defined = self._auxiliary_function_definitions[
                DESCRIPTION_FUNCTION
            ]
            if detection_result.description_defined:
                detection_result.description_output = self._get_description(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.description_exception = err

        try:
            detection_result.reference_defined = self._auxiliary_function_definitions[
                REFERENCE_FUNCTION
            ]
            if detection_result.reference_defined:
                detection_result.reference_output = self._get_reference(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.reference_exception = err

        try:
            detection_result.severity_defined = self._auxiliary_function_definitions[
                SEVERITY_FUNCTION
            ]
            if detection_result.severity_defined:
                detection_result.severity_output = self._get_severity(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.severity_exception = err

        try:
            detection_result.runbook_defined = self._auxiliary_function_definitions[
                RUNBOOK_FUNCTION
            ]
            if detection_result.runbook_defined:
                detection_result.runbook_output = self._get_runbook(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.runbook_exception = err

        try:
            detection_result.destinations_defined = self._auxiliary_function_definitions[
                DESTINATIONS_FUNCTION
            ]
            if detection_result.destinations_defined:
                detection_result.destinations_output = self._get_destinations(
                    event,
                    outputs,
                    outputs_names,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.destinations_exception = err

        try:
            detection_result.dedup_defined = self._auxiliary_function_definitions[DEDUP_FUNCTION]
            if not detection_result.dedup_defined:
                detection_result.dedup_output = self._get_dedup_fallback(
                    detection_result.title_output
                )
            else:
                detection_result.dedup_output = self._get_dedup(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.dedup_exception = err

        try:
            detection_result.alert_context_defined = self._auxiliary_function_definitions[
                ALERT_CONTEXT_FUNCTION
            ]
            if detection_result.alert_context_defined:
                detection_result.alert_context_output = self._get_alert_context(
                    event,
                    use_default_on_exception=batch_mode,
                )
        except Exception as err:  # pylint: disable=broad-except
            detection_result.alert_context_exception = err

        return detection_result

    def _filters_should_run(self) -> bool:
        return (
            # Only running on basic rules for now
            self.detection_type == TYPE_RULE
            and len(self.filters) > 0
            and not self.skip_filters
        )

    def _get_alert_context(
        self, event: Mapping, use_default_on_exception: bool = True
    ) -> Optional[str]:
        try:
            command = getattr(self._module, ALERT_CONTEXT_FUNCTION)
            alert_context = self._run_command(command, event, Mapping)
            serialized_alert_context = json.dumps(alert_context, default=PantherEvent.json_encoder)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                return json.dumps({ALERT_CONTEXT_ERROR_KEY: repr(err)})
            raise

        if len(serialized_alert_context) > MAX_ALERT_CONTEXT_SIZE:
            # If context exceeds max size, return empty one
            alert_context_error = (
                "alert_context size is [{}] characters,"
                " bigger than maximum of [{}] characters".format(
                    len(serialized_alert_context), MAX_ALERT_CONTEXT_SIZE
                )
            )
            return json.dumps({ALERT_CONTEXT_ERROR_KEY: alert_context_error})

        return serialized_alert_context

    # Returns the dedup string for this detection match
    # If the detection match had a custom title, use the title as a deduplication string
    # If no title and no dedup function is defined, return the default dedup string.
    def _get_dedup(
        self,
        event: Mapping,
        use_default_on_exception: bool = True,
    ) -> str:
        try:
            command = getattr(self._module, DEDUP_FUNCTION)
            dedup_string = self._run_command(command, event, str)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "dedup method raised exception. "
                    'Defaulting dedup string to "%s". Exception: %s',
                    self.detection_id,
                    err,
                )
                return self._default_dedup_string
            raise

        if not dedup_string:
            # If dedup string is None or empty, return the default dedup string
            return self._default_dedup_string

        if len(dedup_string) > MAX_DEDUP_STRING_SIZE:
            # If dedup_string exceeds max size, truncate it
            self.logger.info(
                "maximum dedup string size is [%d] characters. "
                "Dedup string for detection with ID [%s] is [%d] characters. Truncating.",
                MAX_DEDUP_STRING_SIZE,
                self.detection_id,
                len(dedup_string),
            )
            num_characters_to_keep = MAX_DEDUP_STRING_SIZE - len(TRUNCATED_STRING_SUFFIX)
            return dedup_string[:num_characters_to_keep] + TRUNCATED_STRING_SUFFIX

        return dedup_string

    def _get_dedup_fallback(self, title: Optional[str]) -> str:
        if title:
            # If no dedup function is defined but the detection
            # had a title, use the title as dedup string
            return title
            # If no dedup function defined, return default dedup string
        return self._default_dedup_string

    def _get_description(
        self, event: Mapping, use_default_on_exception: bool = True
    ) -> Optional[str]:
        try:
            command = getattr(self._module, DESCRIPTION_FUNCTION)
            description = self._run_command(command, event, str)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "description method for detection with id [%s] raised exception. "
                    "Using default Exception: %s",
                    self.detection_id,
                    err,
                )
                return ""
            raise

        if len(description) > MAX_GENERATED_FIELD_SIZE:
            # If generated field exceeds max size, truncate it
            self.logger.info(
                "maximum field [description] length is [%d]. "
                "[%d] for detection with ID [%s] . Truncating.",
                MAX_GENERATED_FIELD_SIZE,
                len(description),
                self.detection_id,
            )
            num_characters_to_keep = MAX_GENERATED_FIELD_SIZE - len(TRUNCATED_STRING_SUFFIX)
            return description[:num_characters_to_keep] + TRUNCATED_STRING_SUFFIX
        return description

    def _get_destinations(  # pylint: disable=too-many-return-statements,too-many-arguments
        self,
        event: Mapping,
        outputs: dict,
        outputs_display_names: dict,
        use_default_on_exception: bool = True,
    ) -> Optional[List[str]]:
        try:
            command = getattr(self._module, DESTINATIONS_FUNCTION)
            destinations = self._run_command(command, event, [])
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info("destinations method raised exception. Exception: %s", err)
                return None
            raise
        # Return early if destinations returned None
        if destinations is None:
            return None

        # Return early if destinations is an empty list (alert dest. suppression)
        if len(destinations) == 0:
            return ["SKIP"]

        # Check for (in)valid destinations
        invalid_destinations = []
        standardized_destinations: List[str] = []

        # Standardize the destinations
        for each_destination in destinations:
            # case for valid display name
            if (
                each_destination in outputs_display_names
                and outputs_display_names[each_destination].destination_id
                not in standardized_destinations
            ):
                standardized_destinations.append(
                    outputs_display_names[each_destination].destination_id
                )
            # case for valid UUIDv4
            elif each_destination in outputs and each_destination not in standardized_destinations:
                standardized_destinations.append(each_destination)
            else:
                invalid_destinations.append(each_destination)

        if invalid_destinations:
            if use_default_on_exception:
                self.logger.info(
                    "destinations method yielded invalid destinations: %s",
                    str(invalid_destinations),
                )
                return None
            raise UnknownDestinationError(
                "Invalid Destinations",
                invalid_destinations,
            )

        if len(standardized_destinations) > MAX_DESTINATIONS_SIZE:
            # If generated field exceeds max size, truncate it
            self.logger.info(
                "maximum len of destinations [%d] for detection with ID "
                "[%s] is [%d] fields. Truncating.",
                MAX_DESTINATIONS_SIZE,
                self.detection_id,
                len(standardized_destinations),
            )
            return standardized_destinations[:MAX_DESTINATIONS_SIZE]
        return standardized_destinations

    def _get_reference(
        self, event: Mapping, use_default_on_exception: bool = True
    ) -> Optional[str]:
        try:
            command = getattr(self._module, REFERENCE_FUNCTION)
            reference = self._run_command(command, event, str)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "reference method for detection with id [%s] raised exception. "
                    "Using default. Exception: %s",
                    self.detection_id,
                    err,
                )
                return ""
            raise

        if len(reference) > MAX_GENERATED_FIELD_SIZE:
            # If generated field exceeds max size, truncate it
            self.logger.info(
                "maximum field [reference] length is [%d]. "
                "[%d] for detection with ID [%s] . Truncating.",
                MAX_GENERATED_FIELD_SIZE,
                len(reference),
                self.detection_id,
            )
            num_characters_to_keep = MAX_GENERATED_FIELD_SIZE - len(TRUNCATED_STRING_SUFFIX)
            return reference[:num_characters_to_keep] + TRUNCATED_STRING_SUFFIX
        return reference

    def _get_runbook(self, event: Mapping, use_default_on_exception: bool = True) -> Optional[str]:
        try:
            command = getattr(self._module, RUNBOOK_FUNCTION)
            runbook = self._run_command(command, event, str)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "runbook method for detection with id [%s] raised exception. "
                    "Using default. Exception: %s",
                    self.detection_id,
                    err,
                )
                return ""
            raise

        if len(runbook) > MAX_GENERATED_FIELD_SIZE:
            # If generated field exceeds max size, truncate it
            self.logger.info(
                "maximum field [runbook] length is [%d]. [%d] for detection with ID [%s]. "
                "Truncating.",
                MAX_GENERATED_FIELD_SIZE,
                len(runbook),
                self.detection_id,
            )
            num_characters_to_keep = MAX_GENERATED_FIELD_SIZE - len(TRUNCATED_STRING_SUFFIX)
            return runbook[:num_characters_to_keep] + TRUNCATED_STRING_SUFFIX
        return runbook

    def _get_severity(self, event: Mapping, use_default_on_exception: bool = True) -> Optional[str]:
        try:
            command = getattr(self._module, SEVERITY_FUNCTION)
            severity = self._run_command(command, event, str).upper()
            if severity not in SEVERITY_TYPES:
                if severity == SEVERITY_DEFAULT:
                    return self.detection_severity
                self.logger.info(
                    "severity method for detection with id [%s] yielded [%s], expected [%s]",
                    self.detection_id,
                    severity,
                    str(SEVERITY_TYPES),
                )
                raise AssertionError(
                    "Expected severity to be any of the following: [%s], got [%s] instead."
                    % (str(SEVERITY_TYPES), severity)
                )
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "severity method for detection with id [%s] raised exception. "
                    "Using default (%s). Exception: %s",
                    self.detection_id,
                    self.detection_severity,
                    err,
                )
                return self.detection_severity
            raise
        return severity

    def _get_title(self, event: Mapping, use_default_on_exception: bool) -> Optional[str]:
        try:
            command = getattr(self._module, TITLE_FUNCTION)
            title = self._run_command(command, event, str)
        except Exception as err:  # pylint: disable=broad-except
            if use_default_on_exception:
                self.logger.info(
                    "title method for detection with id [%s] raised exception. "
                    "Using default. Exception: %s",
                    self.detection_id,
                    err,
                )
                return self.detection_id
            raise

        if len(title) > MAX_GENERATED_FIELD_SIZE:
            # If generated field exceeds max size, truncate it
            self.logger.info(
                "maximum field [title] length is [%d]. "
                "[%d] for detection with ID [%s] . Truncating.",
                MAX_GENERATED_FIELD_SIZE,
                len(title),
                self.detection_id,
            )
            num_characters_to_keep = MAX_GENERATED_FIELD_SIZE - len(TRUNCATED_STRING_SUFFIX)
            return title[:num_characters_to_keep] + TRUNCATED_STRING_SUFFIX
        return title

    def _run_command(self, function: Callable, event: Mapping, expected_type: Any) -> Any:
        if self.suppress_output:
            with contextlib.redirect_stdout(None), contextlib.redirect_stderr(None):
                result = function(event)
        else:
            result = function(event)
        # Branch in case of list
        if not isinstance(expected_type, list):
            if not isinstance(result, expected_type):
                raise FunctionReturnTypeError(
                    "detection [{}] function [{}] returned [{}], expected [{}]".format(
                        self.detection_id,
                        function.__name__,
                        type(result).__name__,
                        expected_type.__name__,
                    )
                )
        else:
            if result is None:
                return result
            if not isinstance(result, list) or not all(isinstance(x, (str, bool)) for x in result):
                raise FunctionReturnTypeError(
                    "detection [{}] function [{}] returned [{}], expected a list".format(
                        self.detection_id, function.__name__, type(result).__name__
                    )
                )
        return result

    def _is_function_defined(self, name: str) -> bool:
        return hasattr(self._module, name)


class Rule(Detection):
    """Panther rule metadata and imported module."""

    # default detection types for rules
    default_detection_type = TYPE_RULE

    # rules have a rule method
    matcher_function_name = "rule"

    # a rule should trigger an alert on True return value
    matcher_alert_value = True

    def matcher_function(self, event: Mapping) -> bool:
        # for scheduled rules the rule function is optional,
        # defaults to True and will pass the events thru
        if self.detection_type == TYPE_SCHEDULED_RULE and not hasattr(
            self._module, self.matcher_function_name
        ):
            return True

        command = getattr(self._module, self.matcher_function_name)
        return self._run_command(command, event, bool)



from distutils.core import setup

VERSION = "0.10.1"

install_requires = [
    'jsonpath-ng',
]

with open('requirements.txt') as f:
    dependencies_with_versions = []
    for dependency in f.readlines():
        dependency_with_version = dependency.strip()
        package_name = dependency_with_version.split('==')[0]
        if package_name in install_requires:
            dependencies_with_versions.append(dependency_with_version)

setup(
    name='panther_core',
    packages=['panther_core', 'panther_core/exec', 'panther_core/snapshots'],
    package_dir={
        'exec': 'panther_core/exec',
        'snapshots': 'panther_core/snapshots',
    },
    version=VERSION,
    license='AGPL-3.0',
    description='Panther core library',
    author='Panther Labs Inc',
    author_email='pypi@runpanther.io',
    url='https://github.com/panther-labs/panther_core',
    download_url=f'https://github.com/panther-labs/panther_core/archive/refs/tags/v{VERSION}.tar.gz',
    keywords=['Security', 'CLI'],
    install_requires=install_requires,
    classifiers=[
        'Topic :: Security',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: GNU Affero General Public License v3',
    ],
)



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from importlib import util as import_util
from pathlib import Path
from typing import Any


def allowed_char(char: str) -> bool:
    """Return true if the character is part of a valid ID."""
    return char.isalnum() or char in {" ", "-", "."}


def id_to_path(directory: str, object_id: str) -> str:
    """Method returns the file path where the module will be stored"""
    safe_id = "".join(x if allowed_char(x) else "_" for x in object_id)
    path = os.path.join(directory, safe_id + ".py")
    return path


def import_file_as_module(path: str, object_id: str) -> Any:
    """Dynamically import a Python module from a file.

    See also: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    """

    spec = import_util.spec_from_file_location(object_id, path)
    mod = import_util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(mod)  # type: ignore
    return mod


def store_modules(path: str, body: str) -> None:
    """Stores modules to disk."""
    # Create dir if it doesn't exist
    Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
    with open(path, "w", encoding="utf-8") as py_file:
        py_file.write(body)


def get_bool_env_var(var_name: str, default: bool = False) -> bool:
    """
    Reads an environment variable as a boolean, returning default if not set.
    Uses direct string comparison for efficiency.
    """
    value = os.getenv(var_name)
    if value is None:
        return default
    return value.lower() in ("true", "1", "yes", "on")



import ipaddress
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Final, Optional

import jmespath

from panther_core import PantherEvent
from panther_core.immutable import ImmutableList

SIMPLE_KEY_REGEX: Final = re.compile(r"^[a-zA-Z0-9_]*$")


def is_list(obj: Any) -> bool:
    """
    Returns True if the given object is, or inherits from:
    - list (Python builtin)
    - ImmutableList (panther-core type)
    """
    return isinstance(obj, (list, ImmutableList))


def is_simple_key(key: str) -> bool:
    """Returns True if the given key is "simple", meaning it can be used
    to look up something in the event dictionary using .get() rather than
    jmespath.

    It considers a key simple if it is composed only of the following.
    - letters
    - numbers
    - underscores

    Any future additions must not conflict with the jmespath specification:
    https://jmespath.org/specification.html
    """
    return re.match(SIMPLE_KEY_REGEX, key) is not None


@dataclass
class FilterConditionBase:
    target: str
    value: Any
    operator: str
    source: Optional[str] = None
    target_type_hint: Optional[str] = None
    compiled_path: Optional[Callable[[Any], Any]] = None

    @classmethod
    def from_dict(cls, data: Dict) -> Any:
        return cls(
            source=data.get("source", None),
            target=data.get("target", ""),
            value=data.get("value", ""),
            operator=data.get("operator", ""),
            target_type_hint=data.get("targetTypeHint"),
        )

    @staticmethod
    def is_valid_dict(data: Dict) -> bool:
        return data.get("target", None) and data.get("operator", None)


class FilterOperator(ABC):  # pylint: disable=too-few-public-methods
    """
    FilterOperator is an abstract class that all operator classes should implement.

    Methods
    -------
    compare(event:Any) -> bool
        this is the main function that will run operation logic. If operation matches compare
        should return true
    """

    @abstractmethod
    def compare(self, event: Any) -> bool:
        pass


@dataclass
class FilterCondition(FilterConditionBase):
    value_is_list: Optional[bool] = None
    value_is_str: Optional[bool] = None
    target_is_ip: Optional[bool] = None

    def __post_init__(self) -> None:
        if is_simple_key(self.target):
            # if target is simple (ie. not a jmespath expression) we can access the
            # property of the event directly, avoiding jmespath overhead
            self.compiled_path = lambda x: x.get(self.target)
        else:
            # use jmespath whenever a more complex expression is detected
            self.compiled_path = jmespath.compile(self.target).search
        self.value_is_list = is_list(self.value)
        self.value_is_str = isinstance(self.value, str)
        self.target_is_ip = self.target_type_hint == "ip"


@dataclass
class BaseFilterOperator(FilterOperator):
    """
    BaseFilterOperator is the base class for all operators.
    It contains the predominant condition field that will be used for all operations.
    """

    condition: FilterCondition

    @abstractmethod
    def compare(self, event: Any) -> bool:
        pass

    def target_value(self, event: Any) -> Any:
        """
        target_value grabs the relevant path defined in the condition and extracts it from
        the provided event
        """
        return self.condition.compiled_path(event)  # type: ignore


_registered_ops: Dict[str, Callable[[FilterCondition], BaseFilterOperator]] = {}


def op(op_id: str) -> Any:  # pylint: disable=invalid-name
    """
    Registers associated operator class with the provided operator.
    Associated class gets instantiated when a comparison is made against a filter condition
    with an operator matching the provided filter operator

    e.g.
        @op('==')
        @dataclass
        class EqualOperator(BaseFilterOperator):
            pass

    Whenever a FilterCondition with the operator of '==' is detected the EqualOperator class will get
    instantiated to validate the condition
    """

    def wrap(cls: Any) -> Any:
        _registered_ops[op_id] = cls
        return cls

    return wrap


@op("==")
@dataclass
class EqualFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        return self.target_value(event) == self.condition.value


@op("!=")
@dataclass
class NotEqualFilterOperator(EqualFilterOperator):
    def compare(self, event: Any) -> bool:
        return not super().compare(event)


@op("<")
@dataclass
class LessThanFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return False
        return target_value < self.condition.value


@op(">")
@dataclass
class GreaterThanFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return False
        return target_value > self.condition.value


@op(">=")
@dataclass
class GreaterThanOrEqualFilterOperator(GreaterThanFilterOperator):
    equal: Optional[EqualFilterOperator] = None

    def __post_init__(self) -> None:
        self.equal = EqualFilterOperator(condition=self.condition)

    def compare(self, event: Any) -> bool:
        return super().compare(event) or self.equal.compare(event)  # type: ignore


@op("<=")
@dataclass
class LessThanOrEqualFilterOperator(LessThanFilterOperator):
    equal: Optional[EqualFilterOperator] = None

    def __post_init__(self) -> None:
        self.equal = EqualFilterOperator(condition=self.condition)

    def compare(self, event: Any) -> bool:
        return super().compare(event) or self.equal.compare(event)  # type: ignore


@op("contains")
@dataclass
class ContainsFilterOperator(BaseFilterOperator):
    """
    The contains operator exhibits distinct behavior with arrays and strings:

    Arrays:
        - Prioritizes exact matches, checking if an element is wholly present within the array.
        - For string arrays specifically, performs partial matches if an exact match is not found.

    Strings:
        - Always conducts partial matches, assessing whether a string is substring within another string.
    """

    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return False
        result = self.condition.value in target_value
        if result:
            return result
        if self.condition.value_is_str and isinstance(target_value, list):
            for string in target_value:
                if self.condition.value in string:
                    return True
        return False


@op("!contains")
@dataclass
class NotContainsFilterOperator(ContainsFilterOperator):
    def compare(self, event: Any) -> bool:
        return not super().compare(event)


@op("in")
@dataclass
class InFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        lhs = self.target_value(event)
        rhs = self.condition.value
        result = False
        if self.condition.value_is_list and is_list(lhs):
            for value in rhs:
                if value in lhs:
                    return True
        else:
            result = lhs in rhs
        return result


@op("!in")
@dataclass
class NotInFilterOperator(InFilterOperator):
    def compare(self, event: Any) -> bool:
        return not super().compare(event)


class FilterException(Exception):
    pass


def to_ip_network(value: Any) -> Any:
    if is_list(value):
        raise FilterException("expected target field to be an ip or cidr but got an array")
    try:
        return ipaddress.ip_network(value, False)
    except ValueError as v:  # pylint: disable=invalid-name
        raise FilterException("invalid ip or cidr detected") from v


@op("in_cidr")
@dataclass
class InCIDRFilterOperator(BaseFilterOperator):
    """
    Expects the event value to be an ip address and expects the condition value to be a CIDR
    e.g.
    event.srcIp IN CIDR 255.255.255.240/24

    would return true if event.srcIp was 255.255.255.2

    note: this works for IPv4 and IPv6 addresses and CIDR blocks
    """

    value_converted: bool = False

    def compare(self, event: Any) -> bool:
        if not self.value_converted:
            self.condition.value = to_ip_network(self.condition.value)
            self.value_converted = True

        target_value = self.target_value(event)
        if target_value is None:
            return False
        lhs = to_ip_network(target_value)
        rhs = self.condition.value

        return lhs.overlaps(rhs)


@op("!in_cidr")
@dataclass
class NotInCIDRFilterOperator(InCIDRFilterOperator):
    def compare(self, event: Any) -> bool:
        return not super().compare(event)


@op("contains_ip")
@dataclass
class ContainsIPFilterOperator(BaseFilterOperator):
    """
    Expects the event value to be a list of ips and expects the condition to be CIDR block
    e.g.
    p_any_ip_addresses contains_ip 255.255.255.0/24

    would return true if p_any_ip_addresses == [255.255.255.2, 127.0.0.1]

    note: this works for IPv4 and IPv6 addresses as well as CIDR blocks
    """

    value_converted: bool = False

    def compare(self, event: Any) -> bool:
        if not self.value_converted:
            self.condition.value = to_ip_network(self.condition.value)
            self.value_converted = True
        lhs = self.target_value(event)
        if lhs is None:
            return False
        if not is_list(lhs):
            raise FilterException(f"expected {self.condition.target} to be a list of ips")

        ips = map(to_ip_network, lhs)
        for ip in ips:  # pylint: disable=invalid-name
            if self.condition.value.overlaps(ip):
                return True

        return False


@op("!contains_ip")
@dataclass
class NotContainsIPFilterOperator(ContainsIPFilterOperator):
    """
    Expects the event value to be a list of ips and expects the condition to be CIDR block
    e.g.
    p_any_ip_addresses not contains_ip 255.255.255.0/24

    would return false if p_any_ip_addresses == [255.255.255.2, 127.0.0.1]

    note: this works for IPv4 and IPv6 addresses and CIDR blocks
    """

    def compare(self, event: Any) -> bool:
        return not super().compare(event)


@op("is_public")
@dataclass
class IsPublicIPFilterOperator(BaseFilterOperator):
    """
    Returns true if the target value is a public IP address

    note: this works for IPv4 and IPv6 addresses
    """

    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return False

        target_ip = ipaddress.ip_address(target_value)
        return target_ip.is_global


@op("is_private")
@dataclass
class IsPrivateIPFilterOperator(BaseFilterOperator):
    """
    Returns true if the target value is a private IP address

    note: this works for IPv4 and IPv6 addresses
    """

    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return False

        target_ip = ipaddress.ip_address(target_value)
        return target_ip.is_private


@op("starts_with")
@dataclass
class StartsWithFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if not hasattr(target_value, "startswith"):
            return False
        return target_value.startswith(self.condition.value)


@op("ends_with")
@dataclass
class EndsWithFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if not hasattr(target_value, "endswith"):
            return False
        return target_value.endswith(self.condition.value)


@op("empty")
@dataclass
class IsEmptyFilterOperator(BaseFilterOperator):
    def compare(self, event: Any) -> bool:
        target_value = self.target_value(event)
        if target_value is None:
            return True
        if hasattr(target_value, "__len__"):
            return len(target_value) == 0
        return False


@op("!empty")
@dataclass
class NotIsEmptyFilterOperator(IsEmptyFilterOperator):
    def compare(self, event: Any) -> bool:
        return not super().compare(event)


class UnsupportedFilterOperator(Exception):
    pass


def get_compare_fn(condition: FilterCondition) -> Any:
    op_cls = _registered_ops.get(condition.operator, None)
    if op_cls is None:
        raise UnsupportedFilterOperator(f"unsupported operator id detected. ({condition.operator})")

    return op_cls(condition).compare


class Filter:  # pylint: disable=too-few-public-methods
    """
    Filter is a recursive data structure that allows filtering by any logical operation against an event.

    e.g.
        (event.srcPort > 443 && event.name IN ["hello", "world"]) || (underAttack == True || !(atPeace == True))

        would be represented as:
        {
            'or': {
                'and': [
                    {
                        'target': 'event.srcPort',
                        'operator': '>',
                        'value': 443
                    },
                    {
                        'target': 'event.name'
                        'operator': 'in',
                        'value': ['hello', 'world']
                    }
                ],
                'or': [
                    {
                        'target': 'underAttack',
                        'operator': '=',
                        'value': True
                    },
                    'not': [
                        {
                            'target': 'atPeace',
                            'operator: '=',
                            'value': True
                        }
                    ]
                ]
            }
        }

    Attributes
    ----------
    condition : Optional[FilterCondition]
        if set will contain a relevant condition
    exception : Optional[BaseException]
        if set the filter will always return with this exception when called
    ands : List[Filter]
        a list of filters if set that will be ANDed together
    ors : List[Filter]
        a list of filters if set that will be ORed together
    nots : List[Filter]
        a list of filters if set that will be ANDed together with a NOT expression

    Methods
    -------
    filter(event:Any)
        this is the main function that will filter a configured filter against an event.
        if an exception occurs during initialization of the filter it will always return the exception/
    """

    condition: Optional[FilterCondition]
    exception: Optional[BaseException] = None

    def filter(self, event: Any) -> bool:
        if self.exception:
            raise self.exception
        val = event
        if isinstance(val, PantherEvent):
            val = event.to_dict()
        return self._compiled_filter(val)

    def total_conditions(self) -> int:
        total = 0
        if self.condition is not None:
            total += 1
        else:
            for field in self.nots:
                total += field.total_conditions()
            for field in self.ands:
                total += field.total_conditions()
            for field in self.ors:
                total += field.total_conditions()

        return total

    def __init__(self, info: Dict):
        self.nots = []
        self.ors = []
        self.ands = []

        for i in info.get("not", []):
            self.nots.append(Filter(i))
        for i in info.get("and", []):
            self.ands.append(Filter(i))
        for i in info.get("or", []):
            self.ors.append(Filter(i))

        if FilterCondition.is_valid_dict(info):
            # FilterCondition.is_valid_dict checks for the target and operation fields.
            # If those are set, this is a target / operator / value filter, so the
            # fields (and, or, not) won't be evaluated, only this condition will be.
            # So, if any of those fields are set, this filter will raise an exception,
            # avoiding potential confusion (otherwise the and/or/not fields would be
            # silently ignored).
            if len(self.ands) > 0 or len(self.ors) > 0 or len(self.nots) > 0:
                self.exception = self._too_many_clauses_exception()
            self.condition = FilterCondition.from_dict(info)
        else:
            if info.get("target") or info.get("operator"):
                self.exception = FilterException(
                    "a condition statement requires both target and operator to be set"
                )
            self.condition = None

        try:
            self._compiled_filter = self.__compile()
        except Exception as e:  # pylint: disable=invalid-name,broad-except
            self.exception = e

    def _compiled_or_fn(self) -> Callable[[Any], bool]:
        def _or(event: Any) -> bool:
            for ops in self.ors:
                if ops.filter(event):
                    return True
            return False

        return _or

    def _compiled_and_fn(self) -> Callable[[Any], bool]:
        def _and(event: Any) -> bool:
            for ops in self.ands:
                if ops.filter(event) is False:
                    return False
            return True

        return _and

    def _compiled_not_fn(self) -> Callable[[Any], bool]:
        def _not(event: Any) -> bool:
            for ops in self.nots:
                if ops.filter(event) is True:
                    return False

            return True

        return _not

    def __compile(self) -> Callable[[Any], bool]:
        if self.condition:
            return get_compare_fn(self.condition)

        if len(self.ors) > 0:
            if len(self.nots) > 0 or len(self.ands) > 0:
                self.exception = self._too_many_clauses_exception()
            return self._compiled_or_fn()

        if len(self.ands) > 0:
            if len(self.nots) > 0:
                self.exception = self._too_many_clauses_exception()
            return self._compiled_and_fn()

        if len(self.nots) > 0:
            return self._compiled_not_fn()

        # If this is an empty filter structure, the filter is a no-op.
        # In detection filtering, this == a True result, since we want to
        # err/fail on the safe side: not excluding an event from alerts.
        return lambda x: True

    def _too_many_clauses_exception(self) -> Exception:
        return FilterException(
            "a statement may set at most one of condition "
            "(target/operator/value), and, or, & not"
        )



"""
Panther Core is a Python library for Panther Detections.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from .exec.results import ExecutionOutput


@dataclass
class TestError:
    """Represents an error caused by any of the functions or a generic one"""

    message: Optional[str] = None


@dataclass
class FunctionTestResult:
    """Defines the result of running a function"""

    # output contains the JSON-encoded return value, None if an error was raised
    output: Optional[str]
    # error contains a TestError instance with the error message or
    # None if no error was raised
    error: Optional[TestError]
    # return value matched expectations
    matched: Optional[bool]

    @classmethod
    def new(
        cls,
        matched: bool,
        output: Optional[Union[bool, str, List[str]]],
        raw_err: Optional[str] = None,
    ) -> Optional["FunctionTestResult"]:
        """Create a new instance while applying
        the necessary transformations to the parameters"""
        if output is None and raw_err is None:
            return None

        if output is not None and not isinstance(output, str):
            output = json.dumps(output)

        return cls(output=output, error=cls.to_test_error(raw_err), matched=matched)

    @staticmethod
    def format_error(err: Optional[str], title: Optional[str] = None) -> Optional[str]:
        """Convert an error string to a structured error message"""
        if err is None:
            return None
        if title is not None:
            prefix = f"{title}: "
        else:
            prefix = ""

        return f"{prefix}{err}"

    @staticmethod
    def to_test_error(err: Optional[str], title: Optional[str] = None) -> Optional[TestError]:
        """Convert an error string to a TestError,
        also properly formatting the error message"""
        if err is None:
            return None
        return TestError(message=FunctionTestResult.format_error(err, title=title))


@dataclass  # pylint: disable=R0902
class TestResultsPerFunction:
    """Container for the results of each function"""

    detectionFunction: Optional[FunctionTestResult]  # pylint: disable=C0103
    titleFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    dedupFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    alertContextFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    descriptionFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    referenceFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    severityFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    runbookFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103
    destinationsFunction: Optional[FunctionTestResult] = None  # pylint: disable=C0103


@dataclass  # pylint: disable=R0902
class TestResult:
    """The structure of the results for a test case evaluation"""

    id: Optional[str]  # pylint: disable=C0103
    name: str
    # The following two fields do not conform to Python's naming
    # conventions, but the field names correspond
    # to response attributes by API & FE.
    # TODO: provide a field name translation step if necessary
    detectionId: Optional[str]  # pylint: disable=C0103
    genericError: Optional[str]  # pylint: disable=C0103
    error: Optional[TestError]
    errored: bool
    passed: bool
    trigger_alert: Optional[bool]
    functions: TestResultsPerFunction


@dataclass
class TestExpectations:
    """Contains the expected values for performing assertions"""

    detection: bool
    # TODO: include assertions for remaining functions, e.g title and alert context


@dataclass
class TestSpecification:
    """The structure of a test case"""

    id: str  # pylint: disable=C0103
    name: str
    data: Dict[str, Any]
    mocks: List[Dict[str, Any]]
    expectations: TestExpectations


# pylint: disable=too-few-public-methods
class TestCaseEvaluator:
    """Translates detection execution results to test case results,
    by performing assertions and determining the status"""

    @classmethod
    def for_policies(
        cls, spec: TestSpecification, exec_output: ExecutionOutput
    ) -> TestCaseEvaluator:
        return cls(spec=spec, exec_output=exec_output, alert_value=False)

    @classmethod
    def for_rules(cls, spec: TestSpecification, exec_output: ExecutionOutput) -> TestCaseEvaluator:
        return cls(spec=spec, exec_output=exec_output, alert_value=True)

    def __init__(self, spec: TestSpecification, exec_output: ExecutionOutput, alert_value: bool):
        if exec_output.details is None:
            raise RuntimeError("TestCaseEvaluator received ExecutionOutput without details")

        self._spec = spec
        self._exec_output = exec_output
        self._exec_details = exec_output.details
        self._exec_match = exec_output.match
        self._detection_alert_value = alert_value

    def _get_result_status(self) -> bool:
        """Get the test status - passing/failing"""
        # Aux functions are executed unconditionally
        # (regardless if the detection matched or not) during testing.
        # Only if the detection is expected to trigger an alert,
        # we want to include errors from other functions in the status.
        if self._spec.expectations.detection == self._detection_alert_value:
            return bool(self._exec_output.trigger_alert and not self._exec_output.errored)

        # expectations match the detection output
        return bool(
            self._spec.expectations.detection
            == self._exec_details.primary_functions.detection.output
        )

    def _get_generic_error_details(self) -> Tuple[Optional[str], Optional[str]]:
        generic_error = None
        generic_error_title = None

        if self._exec_details.input_error is not None:
            generic_error = self._exec_details.input_error
            generic_error_title = "Invalid event"
        elif self._exec_details.setup_error is not None:
            generic_error = self._exec_details.setup_error

        return generic_error, generic_error_title

    def interpret(
        self, ignore_exception_types: Optional[List[Type[Exception]]] = None
    ) -> TestResult:
        """Evaluate the detection result taking into account
        the errors raised during evaluation and
        the test specification expectations"""

        # first, we should update the detection result, taking into account any
        # ignored exception types passed into this test
        function_results = {
            "detectionFunction": FunctionTestResult.new(
                self._spec.expectations.detection
                == self._exec_details.primary_functions.detection.output,
                self._exec_details.primary_functions.detection.output,
                ignoreable_exception(
                    self._exec_details.primary_functions.detection.error, ignore_exception_types
                ),
            )
        }

        # We don't include output from other functions
        # unless the test was expected to match and trigger an alert.
        # If the test fails, providing all the output provides a faster feedback loop,
        # on possible additional failures.
        if self._spec.expectations.detection == self._detection_alert_value:
            function_results.update(
                titleFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.title.error is None,
                    self._exec_details.aux_functions.title.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.title.error, ignore_exception_types
                    ),
                ),
                descriptionFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.description.error is None,
                    self._exec_details.aux_functions.description.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.description.error,
                        ignore_exception_types,
                    ),
                ),
                referenceFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.reference.error is None,
                    self._exec_details.aux_functions.reference.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.reference.error, ignore_exception_types
                    ),
                ),
                severityFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.severity.error is None,
                    self._exec_details.aux_functions.severity.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.severity.error, ignore_exception_types
                    ),
                ),
                runbookFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.runbook.error is None,
                    self._exec_details.aux_functions.runbook.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.runbook.error, ignore_exception_types
                    ),
                ),
                destinationsFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.destinations.error is None,
                    self._exec_details.aux_functions.destinations.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.destinations.error,
                        ignore_exception_types,
                    ),
                ),
                dedupFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.dedup.error is None,
                    self._exec_details.aux_functions.dedup.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.dedup.error, ignore_exception_types
                    ),
                ),
                alertContextFunction=FunctionTestResult.new(
                    self._exec_details.aux_functions.alert_context.error is None,
                    self._exec_details.aux_functions.alert_context.output,
                    ignoreable_exception(
                        self._exec_details.aux_functions.alert_context.error,
                        ignore_exception_types,
                    ),
                ),
            )

        generic_error, generic_error_title = self._get_generic_error_details()

        return TestResult(
            id=self._spec.id,
            name=self._spec.name,
            detectionId=self._exec_match.detection_id if self._exec_match else "",
            genericError=FunctionTestResult.format_error(generic_error, title=generic_error_title),
            errored=self._exec_output.errored,
            error=FunctionTestResult.to_test_error(generic_error, title=generic_error_title),
            # Passing or failing test?
            passed=self._get_result_status(),
            trigger_alert=self._exec_output.trigger_alert,
            functions=TestResultsPerFunction(**function_results),
        )


def ignoreable_exception(
    exception: Any, ignore_exception_types: Optional[List[Type[Exception]]] = None
) -> Optional[str]:
    if ignore_exception_types:
        for exception_type in ignore_exception_types:
            if repr(exception_type) == exception:
                return None
    return exception
